//
// MP3 Sound Decoder V5 AC1 Rev.exp.02
// Copyright(C)'2017- Nagoden / Desktop Station 
//
//アクセサリコマンドで実行できるＭＰ３サウンドデコーダ
//  アドレスはCV1とCV9
//　アドレス　= CV9*256 + CV1
//　アドレス　300の場合 CV9 = 1 CV1 =44　
//  ヘッドライトランドはデジタル入力式　MP3デコーダと同じ機能を実装しています

#include <avr/eeprom.h>
#include <SoftwareSerial.h>
#include "NmraDcc_x.h"
#include "DFPlayer_Mini_Mp3_x.h"


/*************************************************
 *
 * Declaration of Symbols
 *
 *************************************************/

//#define DEBUG      //リリースのときはコメントアウトすること
#define MAN_VER_NUMBER  02   /* Release Ver CV07 */
#define MAN_ID_NUMBER 108  /* Manufacture ID */


#define DECODER_ADDRESS 3
#define MOTOR_PWM_A 11
#define MOTOR_PWM_B 3

#define PIN_LIGHT  A1
#define PIN_HEADF1  A2
#define PIN_HEADF2  A3
#define PIN_HEADD1  A4
#define PIN_HEADD2  A5

#define PIN_AMPUP 5 // アンプボリュームUP
#define PIN_AMPDOWN 6 // アンプボリュームDOWN
#define PIN_BUSY 7

#define CV_MP3_Vol 58             //MP3 Volume
#define CV_AMP_Vol 59             //AMP Volume
#define CV_MP3_F0 110             //F00 Function
#define CV_MP3_F1 111             //F01 Function
#define CV_MP3_F2 112             //F02 Function
#define CV_MP3_01 121             //01th Switch
#define CV_MP3_02 122             //02th Switch
#define CV_MP3_03 123             //03th Switch
#define CV_MP3_04 124             //04th Switch
#define CV_MP3_05 125             //05th Switch
#define CV_MP3_06 126             //06th Switch
#define CV_MP3_07 127             //07th Switch
#define CV_MP3_08 128             //08th Switch
#define CV_MP3_09 129             //09th Switch
#define CV_MP3_10 130             //10th Switch
#define CV_MP3_11 131             //11th Switch
#define CV_MP3_12 132             //12th Switch
#define CV_MP3_13 133             //13th Switch
#define CV_MP3_14 134             //14th Switch
#define CV_MP3_15 135             //15th Switch
#define CV_MP3_16 136             //16th Switch
#define CV_MP3_17 137             //17th Switch
#define CV_MP3_18 138             //18th Switch
#define CV_MP3_19 139             //19th Switch
#define CV_MP3_20 140             //20th Switch

#define WAIT_VOLMP3   20        //MP3 Volumeの通信ウェイト

#define THRESHOLD_MP3PLAY  3     //外部入力のカウントしきい値

/*************************************************
 *
 * Declaration of Classes
 *
 *************************************************/

NmraDcc  Dcc;
DCC_MSG  Packet;

/*************************************************
 *
 * Declaration of Variables
 *
 *************************************************/


//Task Schedule
unsigned long gPreviousL3 = 0;
unsigned long gPreviousL4 = 0;
unsigned long gIntervalRail = 1000;


// CV

//MP3 Control
uint8_t gCV58_MP3_Vol = 26;
uint8_t gMP3NumList[4] = {2, 3, 4, 5};
uint8_t gCVMP3Table[29] = {0};
//ファンクションの変数
uint8_t gSignal[4] = {0};
uint8_t gFuncBits[29] = {0};
//アンプボリューム調整
uint8_t gAmpVol_Up;
uint8_t gAmpVol_Down;


struct CVPair 
{
  uint16_t  CV;
  uint8_t Value;
};

CVPair FactoryDefaultCVs [] = 
{
  {CV_ACCESSORY_DECODER_ADDRESS_LSB, 3},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},
  {CV_29_CONFIG, 0b10000000},
  {CV_MP3_Vol, 26}, //5V PAM must be 28 or less
  {CV_AMP_Vol, 23},
  {CV_MP3_F2 + 0, 2},
  {CV_MP3_F2 + 1, 3},
  {CV_MP3_F2 + 2, 4},
  {CV_MP3_F2 + 3, 5},
  {CV_MP3_01 + 0, 1},
  {CV_MP3_01 + 1, 1},
  {CV_MP3_01 + 2, 1},
  {CV_MP3_01 + 3, 1},
  {CV_MP3_01 + 4, 1},
  {CV_MP3_01 + 5, 1},
  {CV_MP3_01 + 6, 1},
  {CV_MP3_01 + 7, 1},
  {CV_MP3_01 + 8, 1},
  {CV_MP3_01 + 9, 1},
  {CV_MP3_01 + 10, 1},
  {CV_MP3_01 + 11, 1},
  {CV_MP3_01 + 12, 1},
  {CV_MP3_01 + 13, 1},
  {CV_MP3_01 + 14, 1},
  {CV_MP3_01 + 15, 1},
  {CV_MP3_01 + 16, 1},
  {CV_MP3_01 + 17, 1},
  {CV_MP3_01 + 18, 1},
  {CV_MP3_01 + 19, 1},
  
};

/*************************************************
 *
 * Declaration of Functions
 *
 *************************************************/
void(* resetFunc) (void) = 0;  //declare reset function at address 0 
uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);
unsigned long gPreviousWorkTime = 0;
uint16_t gAccessoryAddress = 3;
uint16_t getMyAddr_Acc(void);

void ControlRoomLight(void);
void ControlHeadLight(void);

void Play_VolDown(uint8_t inVolDown);
void Play_VolUp(uint8_t inVolUp);

/*************************************************
 *
 * Functions
 *
 *************************************************/

void notifyCVResetFactoryDefault()
{
  //When anything is writen to CV8 reset to defaults.

  resetCVToDefault();
  
#ifdef DEBUG
  Serial.println("Resetting...");
#endif
  delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
  resetFunc();
}


void resetCVToDefault()
{
  //CVをデフォルトにリセット
#ifdef DEBUG
  Serial.println("CVs being reset to factory defaults");
#endif

  for (int j = 0; j < FactoryDefaultCVIndex; j++ ) {
    Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
  }
}

extern void notifyCVChange( uint16_t CV, uint8_t Value)
{
  //CVが変更されたときのメッセージ
#ifdef DEBUG
  Serial.print("CV ");
  Serial.print(CV);
  Serial.print(" Changed to ");
  Serial.println(Value, DEC);
#endif
}

//**********************************************
//  CVのACK
//**********************************************
void notifyCVAck(void)
{
#ifdef DEBUG
  Serial.println("notifyCVAck");
#endif
//  MOTOR_Ack();
  analogWrite(MOTOR_PWM_B, 0);
  analogWrite(MOTOR_PWM_A, 250);
  delay( 6 );  
  analogWrite(MOTOR_PWM_A, 0);
}
//**************************************************************************
// This function is called whenever a normal DCC Turnout Packet is received
//**************************************************************************
void notifyDccAccState( uint16_t Addr, uint16_t BoardAddr, uint8_t OutputAddr, uint8_t State)
{
  uint16_t aAccAdress = Addr;
#ifdef DEBUG
    Serial.print("notifyDccAccState: ") ;
    Serial.print(Addr,DEC) ;
    Serial.print(",BAdr:");
    Serial.print(BoardAddr,DEC) ;
    Serial.print(",OAdr:");
    Serial.print(OutputAddr,DEC) ;
    Serial.print(", Target Adr:");
    Serial.print(aAccAdress,DEC) ;
    Serial.print(", gAccessoryAddress:");
    Serial.print(gAccessoryAddress,DEC) ;
    Serial.print(", State:");
    Serial.println(State, HEX) ;
#endif
  //---------------------------------------------------------------------
  //アドレスチェック(CVから得た11bit分の信号を比較)　　//151227 fujigay2
  //----------------------------------------------------------------------
  uint8_t aDir = (OutputAddr & 0b001);
   if ((gAccessoryAddress <= aAccAdress) && ((gAccessoryAddress + 19)  >= aAccAdress ))
   {
      if(aDir  == 0)
      {
        PlayMP3_P( aAccAdress - gAccessoryAddress  + 1 );
      }
      else if( aDir == 1)
      {
        PlayMP3_N( aAccAdress - gAccessoryAddress  + 1 );
      }
   }

}  
//********************************************************************************
// アドレスの取得
//********************************************************************************
uint16_t getMyAddr_Acc(void)
{
  uint16_t  Addr ;
  uint8_t   CV29Value ;
  CV29Value = Dcc.getCV( CV_29_CONFIG ) ;
    if( CV29Value & 0b10000000 ){  // Accessory Decoder? 
      Addr = ( Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_LSB ) ;
    }
    else   // Multi-Function Decoder?
    {
      if( CV29Value & 0b00100000 )  // Two Byte Address?
        Addr = ( Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB ) ;

      else
       Addr = Dcc.getCV( 1 ) ;
    }
    
    return Addr ;
}

//**********************************************
//  セットアップ
//**********************************************
void setup()
{
  // ソフトウエアシリアル通信レートセット:
  Serial.begin(9600);
  mp3_set_serial (Serial);  //set softwareSerial for DFPlayer-mini mp3 module

  pinMode(PIN_BUSY, INPUT);  //Mp3 Busy pin 設定
  digitalWrite(PIN_BUSY, HIGH); //Internal pull up enabled
  pinMode( MOTOR_PWM_A, OUTPUT ); // Configure the DCC CV Programing ACK pin for an output
  pinMode( MOTOR_PWM_B, OUTPUT ); // Configure the DCC CV Programing ACK pin for an output

   //ランド端子Ｆ１，Ｆ２，ｂ１，ｂ２の割り当てピン初期化
  pinMode(PIN_LIGHT, OUTPUT);
  digitalWrite(PIN_LIGHT, 0);

  pinMode(PIN_HEADF1, INPUT);
  digitalWrite(PIN_HEADF1, 1);

  pinMode(PIN_HEADF2, INPUT);
  digitalWrite(PIN_HEADF2, 1);

  pinMode(PIN_HEADD1, INPUT);
  digitalWrite(PIN_HEADD1, 1);

  pinMode(PIN_HEADD2, INPUT);
  digitalWrite(PIN_HEADD2, 1);

  //PAM8408アンプ用ボリューム
  pinMode(PIN_AMPUP, OUTPUT);
  pinMode(PIN_AMPDOWN, OUTPUT);
  digitalWrite(PIN_AMPUP, 1);
  digitalWrite(PIN_AMPDOWN, 1);

  
  if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
  {
    //if eeprom has 0xFF then assume it needs to be programmed
#ifdef DEBUG
    Serial.println("CV Defaulting due to blank eeprom");
#endif  
    notifyCVResetFactoryDefault();
   } else{
#ifdef DEBUG
   Serial.println("CV Not Defaulting");
#endif
   }
  // Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
  Dcc.pin(0, 2, 0);

  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_NUMBER, MAN_VER_NUMBER,  FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 0 );

  //Reset task
  gPreviousL3 = millis();
  gPreviousL4 = millis();
  /*  アクセサリアドレス(下位2bitを考慮)の先修得 */
  gAccessoryAddress = getMyAddr_Acc();
  #ifdef DEBUG
  Serial.print("Acc Adr is ");
  Serial.println(gAccessoryAddress);  
  #endif
  
  /* CV 初期化 */
  gCV58_MP3_Vol = Dcc.getCV( CV_MP3_Vol ) ;

  /* アンプボリューム調整 */
  uint8_t aAmpVol = Dcc.getCV( CV_AMP_Vol);
  if ( aAmpVol > 21)
  {
    gAmpVol_Up = aAmpVol - 21;
    gAmpVol_Down = 0;
  }
  else if ( aAmpVol < 21)
  {
    gAmpVol_Up = 0;
    gAmpVol_Down = 21 - aAmpVol;
  }
  else
  {
    //何もしない(Default, 12dB)
    gAmpVol_Up = 0;
    gAmpVol_Down = 0;
  }
  /* MP3ファンクションテーブル初期化・読み取り キーインプット*/
  for( int i = 0; i < 4; i++)
  {
    gMP3NumList[i] = Dcc.getCV( CV_MP3_F2 + i ) ;
  }
  /* MP3ファンクションテーブル初期化・読み取り アクセサリスイッチ*/
  for( int i = 0; i < 20; i++)
  {
    gCVMP3Table[i] = Dcc.getCV( CV_MP3_01 + i ) ;
  }
  
}
//********************************************************
//  ループ処理
//********************************************************
void loop()
{

  // You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
  Dcc.process();
  
  if( (millis() - gPreviousL3) >= 100)
  {
    //入力監視
    TaskSignalPlay();
    
    //Reset task
    gPreviousL3 = millis();   
  }
  
  if ( (millis() - gPreviousL4) >= 250)
  {
     //アンプボリューム調整
    if (gAmpVol_Up > 0)
    {
      Play_VolUp();
      gAmpVol_Up--;
    }

    if ( gAmpVol_Down > 0)
    {
      Play_VolDown();
      gAmpVol_Down--;
    }

    //Reset task

  }
}

//******************************************************
//　ランド入力の再生
//******************************************************
void TaskSignalPlay()
{
  
  if( digitalRead(PIN_HEADF1) == 0)
  {
    gSignal[0]++;
  }
  else
  {
    gSignal[0] = 0;
  }
  
  if( digitalRead(PIN_HEADF2) == 0)
  {
    gSignal[1]++;
  }
  else
  {
    gSignal[1] = 0;
  }
  
  if( digitalRead(PIN_HEADD1) == 0)
  {
    gSignal[2]++;
  }
  else
  {
    gSignal[2] = 0;
  }
  
  if( digitalRead(PIN_HEADD2) == 0)
  {
    gSignal[3]++;
  }
  else
  {
    gSignal[3] = 0;
  }
  
  
  //MP3 Play
  
  for( int i = 0; i < 4; i++)
  {
    
    if( gSignal[i] > THRESHOLD_MP3PLAY )
    {
      PlayMP3_Simple(gMP3NumList[i]);
      gSignal[i] = 0;
    }
  }
}
void exec_function (int function, int pin, int FuncState)
{
  //digitalWrite (pin, FuncState);
}

// *************************************************************************
// MP3 Function
// *************************************************************************
void PlayMP3_P(uint8_t inMusicNo)
{
  PlayMP3_PosiEdge( inMusicNo, &gFuncBits[inMusicNo], gCVMP3Table[inMusicNo - 1]);
}

void PlayMP3_N(uint8_t inMusicNo)
{
  PlayMP3_NegaEdge( inMusicNo, &gFuncBits[inMusicNo], gCVMP3Table[inMusicNo - 1]);
}


void PlayMP3_PosiEdge(uint8_t inMusicNo, uint8_t *iopFnBit, uint8_t inModeCV)
{
  /* 0: ON/OFFのエッジで再生 */
  /* 1: ONレベルで再生(リピート無し) */
  /* 2: ONレベルで再生(リピート有り) */

  if ( *iopFnBit == 0)
  {
    *iopFnBit = 1;
    mp3_set_volume (gCV58_MP3_Vol);
    delay (WAIT_VOLMP3);
    mp3_play (inMusicNo);

    if (inModeCV == 2)
    {
      delay (WAIT_VOLMP3);
      mp3_single_loop (true);
    }
  }
  else
  {
    /* 何もしない */
  }
}

void PlayMP3_NegaEdge(uint8_t inMusicNo, uint8_t *iopFnBit, uint8_t inModeCV)
{

  if ( *iopFnBit == 1)
  {
    *iopFnBit = 0;

    if ((inModeCV == 2) || (inModeCV == 1))
    {
      /* 1: ONレベルで再生(リピート無し) */
      /* 2: ONレベルで再生(リピート有り) */
      mp3_single_loop (false);
      delay (WAIT_VOLMP3);
      mp3_stop () ;
      delay (WAIT_VOLMP3);
    }
    else if (inModeCV == 0)
    {
      /* 0: ON/OFFのエッジで再生 */
      mp3_set_volume (gCV58_MP3_Vol);
      delay (WAIT_VOLMP3);
      mp3_play (inMusicNo) ;
      delay (WAIT_VOLMP3);
    }
    else
    {
      mp3_stop() ;
    }
  }
  else
  {
    //何もしない
  }
}

//void PlayMP3_SetVolume()
//{
//  mp3_set_volume (gCV58_MP3_Vol);
//}


void PlayMP3_Simple(uint8_t num)
{
  //Busyが立っている＝ほかの音がなっているときは鳴らさない。
  if (digitalRead(PIN_BUSY) == HIGH)
  {
    //MP3 Volume Set
    mp3_set_volume (gCV58_MP3_Vol);
    delay(WAIT_VOLMP3);
    mp3_play (num);
  }
}

void Play_VolUp(void)
{
  digitalWrite(PIN_AMPUP, 0);
  delay(33);
  digitalWrite(PIN_AMPUP, 1);
}

void Play_VolDown(void)
{
  digitalWrite(PIN_AMPDOWN, 0);
  delay(33);
  digitalWrite(PIN_AMPDOWN, 1);
}
